require 'set'

module UU
  module OS
    module Util

      # Module providing methods to search for classes matching given criteria.
      module ClassScanner

        # Suppress warning for deprecated constant
        # (else it will be displayed while traversing classes)
        if Object.const_defined?(:Config) && Object.const_defined?(:RbConfig)
          Object.send(:remove_const, :Config)
          Object.send(:const_set, :Config, RbConfig)
        end

        # Find all classes matching criteria evaluated in given assert block.
        # @param [Class] root Class from which to start recursive lookup
        # @param [Proc] assert Block which evaluates whether given class
        #    matches expected criteria. Result of assertion block should
        #    be +true+ or +false+
        # @return [Array<Class>] List of matching classes.
        def find_classes(root = Object, &assert)
          ClassScanner.find_classes(root, &assert)
        end

        # @see #find_classes
        # @private
        def self.find_classes(root = Object, &assert) # :nodoc:
          visited = Set.new
          # On JRuby skip scanning of Java module which causes problems
          # with repeated scan by dynamic creation of proxy classes.
          if Object.const_defined?(:Java)
            visited << Java
          end
          __find_classes(root, visited, &assert)
        end

        # Internal method with level parameter to protect
        # from infinite recursion caused by some modules.
        # @private
        def self.__find_classes(root, visited, &assert) # :nodoc:
          classes = []
          # Skip already visited modules
          return classes if visited.member?(root)
          visited << root
          begin
            constants = root.constants
          rescue ArgumentError
            # On JRuby, constants of some classes cannot be listed causing error
            # raised from get_proxy_or_package_under_package method.
            # We have to ignore these classes.
            return classes
          end
          constants.each do |constant|
            begin
              clazz = root.const_get(constant)
            rescue LoadError, NameError
              # Especially on JRuby, not all objects can be loaded from constant.
              # We have to ignore these objects.
              next
            end
            if clazz.kind_of?(Class)
              classes << clazz if assert.call(clazz)
            elsif clazz.kind_of?(Module)
              classes << clazz if assert.call(clazz)
              classes.unshift(*__find_classes(clazz, visited, &assert))
            end
          end
          return classes
        end
        private_class_method :__find_classes

      end

    end
  end
end
